"""

"""

from bkm10_lib.inputs import BKM10Inputs
from bkm10_lib.cff_inputs import CFFInputs

def validate_configuration(configuration_dictionary: dict, verbose: bool):
    """
    ## Description:
    Validate user's dict of initialization parameters.
    
    ## Parameters:
    configuration_dictionary (dict):
        a dict of user-requests DNN parameters that 
        will be used to attempt to build a TF network.

    verbose (boolean):
        Do you want to see all output of this function evaluation?

    ## Notes
    
    """

    required_keys = [
        "kinematics",
        "cff_inputs",
        "target_polarization", 
        "lepton_beam_polarization"
        ]
    
    for key in required_keys:
        if key not in configuration_dictionary:
            raise ValueError(f"Missing required key in config: {key}")
    
    kinematic_settings = configuration_dictionary["kinematics"]

    if not kinematic_settings:
        raise ValueError("> Missing 'kinematics' key.")

    if not isinstance(kinematic_settings, BKM10Inputs):
        raise TypeError("> 'kinematics' key must be a BKM10Inputs instance.")

    cff_settings = configuration_dictionary["cff_inputs"]

    if not cff_settings:
        raise ValueError("> Missing 'cff_inputs' key.")

    if not isinstance(cff_settings, CFFInputs):
        raise TypeError("> 'cff_inputs' key must be a CFFInputs instance.")
    
    target_polarization = configuration_dictionary["target_polarization"]

    if target_polarization is None:
        raise ValueError("> Missing 'target_polarization' key.")

    if target_polarization != 0. and target_polarization != 0.5 and target_polarization != -0.5:
        raise TypeError("> 'target_polarization' key must be a *float* of -1.0, 0.0, or 1.0 only.")
    
    lepton_beam_polarization = configuration_dictionary["lepton_beam_polarization"]

    if lepton_beam_polarization is None:
        raise ValueError("> Missing 'lepton_beam_polarization' key.")

    if lepton_beam_polarization != 0. and lepton_beam_polarization != 1.0 and lepton_beam_polarization != -1.0:
        raise TypeError("> 'lepton_beam_polarization' key must be a *float* of -0.5, 0.0, or 0.5 only.")
    
    return configuration_dictionary


"""
Entry point for `DifferentialCrossSection` class.
"""

# (X): Import native libraries | shutil
import shutil

# (X): Import native libraries | warnings:
import warnings

# (X): Import third-party libraries | NumPy:
import numpy as np

# (X): Import third-party libraries | Matplotlib:
import matplotlib.pyplot as plt

# (X): 
from bkm10_lib import backend

# (X): Import accompanying modules | bkm10_lib > validation > validate_configuration
from bkm10_lib.validation import validate_configuration

# (X): Import accompanying modules | bkm10_lib > formalism > BKMFormalism:
from bkm10_lib.formalism import BKMFormalism

class DifferentialCrossSection:
    """
    Welcome to the `DifferentialCrossSection` class!

    ## Description:
    Compute BKM10 differential cross sections using user-defined inputs.

    ## Parameters
    configuration : dict
        A dictionary containing the configuration settings with the following keys:
        
        - "kinematics" : BKM10Inputs
            Dataclass containing the required kinematic variables.
        
        - "cff_inputs" : Any
            Object or dictionary containing Compton Form Factor values or parameters.
        
        - "target_polarization" : float
            Polarization value for the target (e.g., 0 for unpolarized).
        
        - "lepton_beam_polarization" : float
            Polarization of the lepton beam (e.g., +1 or -1).

    verbose : bool
        A boolean flag that will tell the class to print out various messages at
        intermediate steps in the calculation. Useful if you want to determine when
        you have, say, calculated a given coefficient, like C_{++}^{LP}(n = 1).
    
    debugging : bool
        A boolean flag that will bomb anybody's terminal with output. As the flag is
        entitled, DO NOT USE THIS unless you need to do some serious debugging. We are
        talking about following how the data gets transformed through every calculation.
    """

    def __init__(self, configuration = None, verbose = False, debugging = False):
        """
        ## Description:
        Initialize the class!

        ## Parameters:

            configuration: (dict)
                A dictionary of configuration parameters

            verbose: (bool)
                Boolean setting to turn on if you want to see 
                frequent print output that shows you "where" the code  
                is in its execution.

            deugging: (bool)
                Do not turn this on.
        """
        
        # (X): Obtain a True/False to operate the calculation in:
        self.configuration_mode = configuration is not None

        # (X): Determine verbose mode:
        self.verbose = verbose

        # (X): Determine debugging mode (DO NOT TURN ON!):
        self.debugging = debugging

        # (X): A dictionary of *every coefficient* that we computed:
        self.coefficients = {}

        # (X): The Trento Angle convention basically shifts all phi to pi - phi:
        self._using_trento_angle_convention = True

        # (X): Hidden data that says if configuration passed:
        self._passed_configuration = False

        # (X): Hidden data that tells us if the functions executed correctly:
        self._evaluated = False

        if self.verbose:
            print("> [VERBOSE]: Verbose mode on.")
        if self.debugging:
            print("> [DEBUGGING]: Debugging mode is on — DO NOT USE THIS!")

        if configuration:
            if self.verbose:
                print("> [VERBOSE]: Configuration dictionary received!")
            if self.debugging:
                print("> [DEBUGGING]:Configuration dictionary received:\n{configuration}")

            try:
                if self.debugging:
                    print("> [DEBUGGING]: Trying to initialize configuration...")
            
                # (X): Initialize the class from the dictionary:
                self._initialize_from_config(configuration)

                if self.debugging:
                    print("> [DEBUGGING]: Configuration passed!")

            except:
                raise Exception("> Unable to initialize configuration!")
            
            self._passed_configuration = True

            if self.verbose:
                print("> [VERBOSE]: Configuration succeeded!")
            if self.debugging:
                print(f"> [DEBUGGING]: Configuration succeeded! Now set internal attribute: {self._passed_configuration}")

    @staticmethod
    def _set_plot_style():
        """
        ## Description:
            We want the plots to look a particular way. So, let's do that.
            In particular, we check if a LaTeX distribution is installed!
        """
        
        # (X): Call shutil to find a TeX distribution:
        latex_installed = shutil.which("latex") is not None

        # (X): If TeX was found...
        if latex_installed:

            # (X): ... matplotlib will not crash if we put this in our plots:
            plt.rcParams.update({
                "text.usetex": True,
                "font.family": "serif",
                "text.latex.preamble": r"\usepackage{amsmath}"
            })

        # (X): If TeX was not found...
        else:

            # (X): ... first, tell the user that we recommend a TeX distribution:
            warnings.warn(
                "> LaTeX is not installed. Falling back to Matplotlib's mathtext.",
                UserWarning)
            
            # (X): If we don't have TeX, then we have to set it to false:
            plt.rcParams.update({
                "text.usetex": False,
                "font.family": "serif"
            })
    
        # (X): Set the rest of the rcParams:
        plt.rcParams['xtick.direction'] = 'in'
        plt.rcParams['xtick.major.size'] = 5
        plt.rcParams['xtick.major.width'] = 0.5
        plt.rcParams['xtick.minor.size'] = 2.5
        plt.rcParams['xtick.minor.width'] = 0.5
        plt.rcParams['xtick.minor.visible'] = True
        plt.rcParams['xtick.top'] = True
        plt.rcParams['ytick.direction'] = 'in'
        plt.rcParams['ytick.major.size'] = 5
        plt.rcParams['ytick.major.width'] = 0.5
        plt.rcParams['ytick.minor.size'] = 2.5
        plt.rcParams['ytick.minor.width'] = 0.5
        plt.rcParams['ytick.minor.visible'] = True
        plt.rcParams['ytick.right'] = True

    def _initialize_from_config(self, configuration_dictionary: dict):
        """
        ## Description:
        We demand a dictionary type, extract each of its keys and values, and then
        we perform validation on each of the values. These values *cannot* be anything!
        So, this function is responsible for that.
        """

        # (1): Initialize a try-catch block:
        try:

            # (1.1): Pass the whole dictionary into the validation function:
            validated_configuration_dictionary = validate_configuration(configuration_dictionary, self.verbose)

            # (1.2): If validation is passed, we *set* the kinematic inputs using the `kinematics` key.
            # | Should be of type `BKMInputs`!
            self.kinematic_inputs = validated_configuration_dictionary["kinematics"]

            # (1.3): Assuming (1.2) passed, then we continue to extract dictionary keys.
            # | Here, it is `cff_inputs`, and should be of type `CFFInputs`.
            self.cff_inputs = validated_configuration_dictionary["cff_inputs"]

            # (1.4): Extract the (float) target polarization.
            self.target_polarization = validated_configuration_dictionary["target_polarization"]

            # (1.5): Extract the (float) lepton polarization.
            self.lepton_polarization = validated_configuration_dictionary["lepton_beam_polarization"]

            # (1.6): Extract the boolean value that tells us to evaluate with/out the WW relations:
            self.using_ww = validated_configuration_dictionary["using_ww"]

            # (1.7): Initialize a BKM formalism with lepton polarization = +1.0.
            # | We do this for unpolarized beams, i.e. when lambda = 0.0:
            self.formalism_plus = self._build_formalism_with_polarization(+1.0)
            
            # (1.8): Initialize a BKM formalims with lepton polarization = -1.0:
            self.formalism_minus = self._build_formalism_with_polarization(-1.0)

        # (2): If there are errors in the initialization above...
        except Exception as error:

            # (2.1): ... too general, yes, but not sure what we put here yet:
            raise Exception("> Error occurred during validation...") from error
        
    def _build_formalism_with_polarization(self, lepton_polarization: float) -> BKMFormalism:
        """
        ## Description:
        There was not an easy way to handle unpolarized lepton beams using what we originally
        coded. The best way to handle it is to actually instantiate *two* separate classes with 
        the two different lepton polarizations. Then, we can easily calculate 
        0.5 * (dsigma(+1) + dsigma(-1)), which is an cross-section for an unpolarized beam.
        """
        # (1): Immediately return a BKMFormalism instance:
        return BKMFormalism(
            inputs = self.kinematic_inputs,
            cff_values = self.cff_inputs,
            lepton_polarization = lepton_polarization,
            target_polarization = self.target_polarization,
            using_ww = self.using_ww,
            verbose = self.verbose,
            debugging = self.debugging)
        
    def compute_prefactor(self) -> float:
        """
        ## Description:
        Immediately compute the prefactor that multiplies the
        entire cross section:
        """
        return self.formalism_plus.compute_cross_section_prefactor()
        
    def compute_c0_coefficient(self, phi_values):
        """
        ## Description:
        We compute the coefficient so-called $c_{0}$ in the mode expansion
        as according to the BKM10 formalism.
        """
        # if not hasattr(self, "formalism"):
        #     raise RuntimeError("> Formalism not initialized. Make sure configuration is valid.")

        if self.lepton_polarization == 0.:
            return 0.5 * (self.formalism_plus.compute_c0_coefficient(phi_values) + self.formalism_minus.compute_c0_coefficient(phi_values))

        elif self.lepton_polarization == 1.0:
            return self.formalism_plus.compute_c0_coefficient(phi_values)
        
        elif self.lepton_polarization == -1.0:
            return self.formalism_minus.compute_c0_coefficient(phi_values)
    
    def compute_c1_coefficient(self, phi_values):
        """
        ## Description:
        We compute the coefficient so-called $c_{1}$ in the mode expansion
        as according to the BKM10 formalism.
        """
        # if not hasattr(self, "formalism"):
        #     raise RuntimeError("> Formalism not initialized. Make sure configuration is valid.")

        if self.lepton_polarization == 0.:
            return 0.5 * (self.formalism_plus.compute_c1_coefficient(phi_values) + self.formalism_minus.compute_c1_coefficient(phi_values))

        elif self.lepton_polarization == 1.0:
            return self.formalism_plus.compute_c1_coefficient(phi_values)
        
        elif self.lepton_polarization == -1.0:
            return self.formalism_minus.compute_c1_coefficient(phi_values)
    
    def compute_c2_coefficient(self, phi_values):
        """
        ## Description:
        We compute the coefficient so-called $c_{2}$ in the mode expansion
        as according to the BKM10 formalism.
        """
        # if not hasattr(self, "formalism"):
        #     raise RuntimeError("> Formalism not initialized. Make sure configuration is valid.")

        if self.lepton_polarization == 0.:
            return 0.5 * (self.formalism_plus.compute_c2_coefficient(phi_values) + self.formalism_minus.compute_c2_coefficient(phi_values))

        elif self.lepton_polarization == 1.0:
            return self.formalism_plus.compute_c2_coefficient(phi_values)
        
        elif self.lepton_polarization == -1.0:
            return self.formalism_minus.compute_c2_coefficient(phi_values)
    
    def compute_c3_coefficient(self, phi_values):
        """
        ## Description:
        We compute the coefficient so-called $c_{3}$ in the mode expansion
        as according to the BKM10 formalism.
        """
        # if not hasattr(self, "formalism"):
        #     raise RuntimeError("> Formalism not initialized. Make sure configuration is valid.")

        if self.lepton_polarization == 0.:
            return 0.5 * (self.formalism_plus.compute_c3_coefficient(phi_values) + self.formalism_minus.compute_c3_coefficient(phi_values))

        elif self.lepton_polarization == 1.0:
            return self.formalism_plus.compute_c3_coefficient(phi_values)
        
        elif self.lepton_polarization == -1.0:
            return self.formalism_minus.compute_c3_coefficient(phi_values)
    
    def compute_s1_coefficient(self, phi_values):
        """
        ## Description:
        We compute the coefficient so-called $s_{1}$ in the mode expansion
        as according to the BKM10 formalism.
        """
        # if not hasattr(self, "formalism"):
        #     raise RuntimeError("> Formalism not initialized. Make sure configuration is valid.")

        if self.lepton_polarization == 0.:
            return 0.5 * (self.formalism_plus.compute_s1_coefficient(phi_values) + self.formalism_minus.compute_s1_coefficient(phi_values))

        elif self.lepton_polarization == 1.0:
            return self.formalism_plus.compute_s1_coefficient(phi_values)
        
        elif self.lepton_polarization == -1.0:
            return self.formalism_minus.compute_s1_coefficient(phi_values)
        
    def compute_s2_coefficient(self, phi_values):
        """
        ## Description:
        We compute the coefficient so-called $s_{2}$ in the mode expansion
        as according to the BKM10 formalism.
        """
        # if not hasattr(self, "formalism"):
        #     raise RuntimeError("> Formalism not initialized. Make sure configuration is valid.")

        if self.lepton_polarization == 0.:
            return 0.5 * (self.formalism_plus.compute_s2_coefficient(phi_values) + self.formalism_minus.compute_s2_coefficient(phi_values))

        elif self.lepton_polarization == 1.0:
            return self.formalism_plus.compute_s2_coefficient(phi_values)
        
        elif self.lepton_polarization == -1.0:
            return self.formalism_minus.compute_s2_coefficient(phi_values)
    
    def compute_s3_coefficient(self, phi_values):
        """
        ## Description:
        We compute the coefficient so-called $s_{3}$ in the mode expansion
        as according to the BKM10 formalism.
        """
        # if not hasattr(self, "formalism"):
        #     raise RuntimeError("> Formalism not initialized. Make sure configuration is valid.")
        
        if self.lepton_polarization == 0.:
            return 0.5 * (self.formalism_plus.compute_s3_coefficient(phi_values) + self.formalism_minus.compute_s3_coefficient(phi_values))

        elif self.lepton_polarization == 1.0:
            return self.formalism_plus.compute_s3_coefficient(phi_values)
        
        elif self.lepton_polarization == -1.0:
            return self.formalism_minus.compute_s3_coefficient(phi_values)

    def compute_cross_section(self, phi_values):
        """
        ## Description:
        We compute the four-fold *differential cross-section* as 
        described with the BKM10 Formalism.

        ## Arguments:
        
        phi: backend.math.ndarray
            A NumPy array that will be plugged-and-chugged into the BKM10 formalism.
        """

        # (X): If  the user has not filled in the class inputs...
        if not hasattr(self, 'kinematic_inputs'):

            # (X): ...enforce the class to use this setting:
            raise RuntimeError("> Missing 'kinematic_inputs' configuration before evaluation.")

        # (X): If the user wants some confirmation that stuff is good...
        if self.verbose:

            # (X): ... we simply evaluate the length of the phi array:
            print(f"> [VERBOSE]: Evaluating cross-section at {len(phi_values)} phi points.")

        # (X): If the user wants to see everything...
        if self.debugging:

            # (X): ... we give it to them:
            print(f"> [DEBUGGING]: Evaluating cross-section with phi values of:\n> {phi_values}")

        # (X): Remember what the Trento angle convention is...
        if self._using_trento_angle_convention:

            # (X): ...if it's on, we apply the shift to the angle array:
            verified_phi_values = backend.math.pi - backend.math.atleast_1d(phi_values)

        # (X): Otherwise...
        else:

            # (X): ... just verify that the array of angles is at least 1D:
            verified_phi_values = backend.math.atleast_1d(phi_values)

        # (X): Obtain the cross-section prefactor:
        cross_section_prefactor = self.compute_prefactor()

        # (X): Obtain coefficients:
        coefficient_c_0 = self.compute_c0_coefficient(verified_phi_values)
        coefficient_c_1 = self.compute_c1_coefficient(verified_phi_values)
        coefficient_c_2 = self.compute_c2_coefficient(verified_phi_values)
        coefficient_c_3 = self.compute_c3_coefficient(verified_phi_values)
        coefficient_s_1 = self.compute_s1_coefficient(verified_phi_values)
        coefficient_s_2 = self.compute_s2_coefficient(verified_phi_values)
        coefficient_s_3 = self.compute_s3_coefficient(verified_phi_values)

        # (X): Compute the dfferential cross-section:
        differential_cross_section = (.389379 * 1000000. * (cross_section_prefactor * (
            coefficient_c_0 * backend.math.cos(0. * verified_phi_values) +
            coefficient_c_1 * backend.math.cos(1. * verified_phi_values) +
            coefficient_c_2 * backend.math.cos(2. * verified_phi_values) +
            coefficient_c_3 * backend.math.cos(3. * verified_phi_values) +
            coefficient_s_1 * backend.math.sin(1. * verified_phi_values) +
            coefficient_s_2 * backend.math.sin(2. * verified_phi_values) +
            coefficient_s_3 * backend.math.sin(3. * verified_phi_values))))
        
        # (X): Store cross-section data as class attribute:
        self.cross_section_values = differential_cross_section

        # (X): The class has now evaluated:
        self._evaluated = True

        # (X): Return the cross section:
        return differential_cross_section
    
    def compute_bsa(self, phi_values):
        """
        ## Description:
        We compute the BKM-predicted BSA.

        ## Arguments:
        
        phi: backend.math.ndarray
            A NumPy array that will be plugged-and-chugged into the BKM10 formalism.
        """

        # (X): If  the user has not filled in the class inputs...
        if not hasattr(self, 'kinematic_inputs'):

            # (X): ...enforce the class to use this key:
            raise RuntimeError("> Missing 'kinematic_inputs' configuration before evaluation.")

        # (X): If the user wants some confirmation that stuff is good...
        if self.verbose:

            # (X): ... we simply evaluate the length of the phi array:
            print(f"> [VERBOSE]: Evaluating cross-section at {len(phi_values)} phi points.")

        # (X): If the user wants to see everything...
        if self.debugging:

            # (X): ... we give it to them:
            print(f"> [DEBUGGING]: Evaluating cross-section with phi values of:\n> {phi_values}")

        # (X): Remember what the Trento angle convention is...
        if self._using_trento_angle_convention:

            # (X): ...if it's on, we apply the shift to the angle array:
            verified_phi_values = backend.math.pi - backend.math.atleast_1d(phi_values)

        # (X): Otherwise...
        else:

            # (X): ... just verify that the array of angles is at least 1D:
            verified_phi_values = backend.math.atleast_1d(phi_values)

        # (X): Compute the differential cross-section according to lambda = +1.0:
        sigma_plus = (
            self.formalism_plus.compute_c0_coefficient(verified_phi_values) * backend.math.cos(0. * verified_phi_values)
            + self.formalism_plus.compute_c1_coefficient(verified_phi_values) * backend.math.cos(1. * verified_phi_values)
            + self.formalism_plus.compute_c2_coefficient(verified_phi_values) * backend.math.cos(2. * verified_phi_values)
            + self.formalism_plus.compute_c3_coefficient(verified_phi_values) * backend.math.cos(3. * verified_phi_values)
            + self.formalism_plus.compute_s1_coefficient(verified_phi_values) * backend.math.sin(1. * verified_phi_values)
            + self.formalism_plus.compute_s2_coefficient(verified_phi_values) * backend.math.sin(2. * verified_phi_values)
            + self.formalism_plus.compute_s3_coefficient(verified_phi_values) * backend.math.sin(3. * verified_phi_values)
            )
    
        # (X): Compute the differential cross-section according to lambda = +1.0:
        sigma_minus = (
            self.formalism_minus.compute_c0_coefficient(verified_phi_values) * backend.math.cos(0. * verified_phi_values)
            + self.formalism_minus.compute_c1_coefficient(verified_phi_values) * backend.math.cos(1. * verified_phi_values)
            + self.formalism_minus.compute_c2_coefficient(verified_phi_values) * backend.math.cos(2. * verified_phi_values)
            + self.formalism_minus.compute_c3_coefficient(verified_phi_values) * backend.math.cos(3. * verified_phi_values)
            + self.formalism_minus.compute_s1_coefficient(verified_phi_values) * backend.math.sin(1. * verified_phi_values)
            + self.formalism_minus.compute_s2_coefficient(verified_phi_values) * backend.math.sin(2. * verified_phi_values)
            + self.formalism_minus.compute_s3_coefficient(verified_phi_values) * backend.math.sin(3. * verified_phi_values)
            )

        # (X): Compute the numerator of the BSA: sigma(+) - sigma(-):
        numerator = sigma_plus - sigma_minus

        # (X): Compute the denominator of the BSA: sigma(+) + sigma(-):
        denominator = sigma_plus + sigma_minus

        # (X): Compute the dfferential cross-section:
        bsa_values = numerator / denominator
        
        # (X): Store cross-section data as class attribute:
        self.bsa_values = bsa_values

        # (X): Return the cross section:
        return bsa_values
    
    def get_coefficient(self, name: str):
        """
        ## Description:
        An interface to query a given BKM coefficient
        """

        # (X): ...
        if not self._evaluated:

            # (X): ...
            raise RuntimeError("Call `evaluate(phi)` first before accessing coefficients.")
        
        # (X): In case there is an issue:
        try:
            
            # (X): Return the coefficient:
            return self.coefficients.get(name, None)
        
        # (X): Catch general exceptions:
        except Exception as exception:

            # (X): Raise an error:
            raise NotImplementedError(f"> Something bad happened...: {exception}")
        
    def plot_cross_section(self, phi_values):
        """
        ## Description:
        Plot the four-fold differential cross-section as a function of azimuthal angle φ.

        ## Arguments:
        phi_values : backend.math.ndarray
            Array of φ values (in degrees) at which to compute and plot the cross-section.
        """

        # (X): We need to check if the cross-section has been evaluated yet:
        if not self._evaluated:
            if self.verbose:
                print("> [VERBOSE]: No precomputed cross-section found. Computing now...")
            if self.debugging:
                print("> [DEBUGGING]: No precomputed cross-section found. Computing now...")

            self.cross_section_values = self.compute_cross_section(phi_values)

        else:
            if self.verbose:
                print("> [VERBOSE]: Found cross-section data... Now constructing plots.")

        self._set_plot_style()

        cross_section_figure_instance, cross_section_axis_instance = plt.subplots(figsize = (8, 5))

        cross_section_axis_instance.plot(phi_values, self.cross_section_values, color = 'black')
        cross_section_axis_instance.set_xlabel(r"Azimuthal Angle $\phi$ (degrees)", fontsize = 14)
        cross_section_axis_instance.set_ylabel(r"$\frac{d^4\sigma}{dQ^2 dx_B dt d\phi}$ (nb/GeV$^4$)", fontsize = 14)
        cross_section_axis_instance.grid(True)
        # cross_section_axis_instance.legend(fontsize = 12)

        try:
            kinematics = self.kinematic_inputs

            title_string = (
                rf"$Q^2 = {kinematics.squared_Q_momentum_transfer:.2f}$ GeV$^2$, "
                rf"$x_B = {kinematics.x_Bjorken:.2f}$, "
                rf"$t = {kinematics.squared_hadronic_momentum_transfer_t:.2f}$ GeV$^2$, "
                rf"$k = {kinematics.lab_kinematics_k:.2f}$ GeV"
                )
            
            cross_section_axis_instance.set_title(title_string, fontsize = 14)

        except AttributeError:

            if self.verbose:
                print("> [VERBOSE]: Could not find full kinematics for title.")

            cross_section_axis_instance.set_title(r"Differential Cross Section vs. $\phi$", fontsize = 14)

        plt.tight_layout()
        plt.show()

    def plot_bsa(self, phi_values):
        """
        ## Description:
        Plot the BKM-predicted BSA with azimuthal angle φ.

        ## Arguments:
        phi_values : backend.math.ndarray
            Array of φ values (in degrees) at which to compute and plot the cross-section.
        """

        # (X): We need to check if the cross-section has been evaluated yet:
        if not self._evaluated:
            if self.verbose:
                print("> [VERBOSE]: No precomputed cross-section found. Computing now...")

            if self.debugging:
                print("> [DEBUGGING]: No precomputed cross-section found. Computing now...")

            self.bsa_values = self.compute_bsa(phi_values)

        else:
            if self.verbose:
                print("> [VERBOSE]: Found cross-section data... Now constructing plots.")

        self._set_plot_style()

        bsa_figure_instance, bsa_axis_instance = plt.subplots(figsize = (8, 5))

        bsa_axis_instance.plot(
            phi_values,
            self.bsa_values,
            color = 'black')
        bsa_axis_instance.set_xlabel(
            r"Azimuthal Angle $\phi$ (degrees)",
            fontsize = 14)
        bsa_axis_instance.set_ylabel(
            r"$\frac{d^4\sigma \left( \lambda = +1 \right) - d^4\sigma \left( \lambda = -1 \right)}{d^4\sigma \left( \lambda = +1 \right) + d^4\sigma \left( \lambda = -1 \right)}$ (unitless)",
            fontsize = 14)
        bsa_axis_instance.grid(True)

        try:
            kinematics = self.kinematic_inputs

            title_string = (
                rf"$Q^2 = {kinematics.squared_Q_momentum_transfer:.2f}$ GeV$^2$, "
                rf"$x_B = {kinematics.x_Bjorken:.2f}$, "
                rf"$t = {kinematics.squared_hadronic_momentum_transfer_t:.2f}$ GeV$^2$, "
                rf"$k = {kinematics.lab_kinematics_k:.2f}$ GeV"
                )
            
            bsa_axis_instance.set_title(f"BSA for {title_string}", fontsize = 14)

        except AttributeError:

            if self.verbose:
                print("> [VERBOSE]: Could not find full kinematics for title.")

            bsa_axis_instance.set_title(r"BSA vs. $\phi$", fontsize = 14)

        plt.tight_layout()
        plt.show()


"""
The BKM10 formalism has been needed for machine-learning purposes.
Here, we provide the logic that facilitates usage of the library
in TensorFlow contexts.
"""

# 3rd Party Library | NumPy:
import numpy as _np

# 3rd Party Library | TensorFlow:
import tensorflow as _tf

# (X): Set the default backend "math computation" library to be NumPy:
_backend = "numpy"

def set_backend(backend):
    """
    ## Description:
    We provide an interface for the user to inform the library
    what kind of math computation is required.
    """

    # (X): In order for this to work, we need to make the backend option global:
    global _backend

    # (X): If a provided backend value does not correspond to NumPy or TensorFlow..
    if backend not in ["numpy", "tensorflow"]:

        # (X): ... raise a Value error demanding the *value* be NP or TF:
        raise ValueError("Backend must be 'numpy' or 'tensorflow'")
    
    # (X): Otherwise, we go ahead and set the backend to whatever it is:
    _backend = backend

def get_backend():
    """
    ## Description:
    Now that we have provided the opportunity change the backend setting,
    we need to be able to find what it is currently set to.
    """

    # (X): Return the backend variable:
    return _backend

def safe_cast(x, promote_to_complex_if_needed=False):
    """
    ## Description:
    In order to handle `complex` types *and* TensorFlow tensors,
    we need to implement a function that will *always* guarantee that
    we are not performing illegal operations between Tensors of different
    type
    - In TensorFlow mode, floats become tf.float32, complexes become tf.complex64.
    - If promote_to_complex_if_needed is True, tf.float32 values are promoted to tf.complex64.
    """
    if get_backend() == "tensorflow":
        if isinstance(x, float):
            x = _tf.constant(x, dtype = _tf.float32)

        elif isinstance(x, complex):
            x = _tf.complex(_tf.constant(x.real, dtype = _tf.float32),
                           _tf.constant(x.imag, dtype = _tf.float32))
            
        elif not _tf.is_tensor(x):
            x = _tf.convert_to_tensor(x)

        if promote_to_complex_if_needed and x.dtype == _tf.float32:
            x = _tf.complex(x, _tf.zeros_like(x))

    return x

class MathWrapper:
    """
    ## Description:
    The `MathWrapper` class helps us easily pass between computing
    values with NumPy or TensorFlow.
    """
    def __getattr__(self, name):
        """
        ## Description:
        In order to handle equivalent math operations between NumPy and
        TensorFlow, we need to be able to *map* (bjectively) between the
        two libraries. So, we need this wrapper to handle this check: 
        Whenever a "math attribute" is used, we have to check if we're
        going to use NumPy or TensorFlow based on the backend setting and
        then evaluate it accordingly.

        ## Notes:
        The main reason we have this --- perhaps the *only* reason --- is 
        because raising things to powers in NumPy is done with np.power() but
        it is tf.pow() in TensorFlow... Thanks, Obama!
        """

        # (X): Sly trick expose either NP or TF computing based on backkend setting:
        mod = _np if _backend == "numpy" else _tf

        # (X): The major exception is that np.power() is equivalent to tf.pow()... Annoying!
        if name == "power" and _backend == "tensorflow":
            return _tf.pow
       
        # (X): Another exception is that the constant pi is *not* in TensorFlow. Yay!
        if name == "pi":
            return _np.pi if _backend == "numpy" else _tf.constant(_np.pi, dtype = _tf.float32)
        
        # (X): TensorFlow also doesn't come with Euler's Number:
        if name == "e":
            return _np.e if _backend == "numpy" else _tf.constant(_np.e, dtype=_tf.float32)
        
        # (X): A major, major hurdle is having TensorFlow handle complex types!
        if name == "complex":
            return complex if _backend == "numpy" else lambda real_part, imag_part: _tf.complex(real_part, imag_part)
        
        # (X): Since we're relying on complex types, NumPy and TensorFlow will not handle them similarly,
        # | so we need to tell it when to us `.real` versus TensorFlow's `.real`.
        if name == "real":
            return (lambda x: x.real) if _backend == "numpy" else _tf.math.real
        
        # (X): Same as above, but for `.imag`:
        if name == "imag":
            return (lambda x: x.imag) if _backend == "numpy" else _tf.math.imag
        
        # (X): We also used the NumPy method `atleast_1d`
        if name == "atleast_1d":

            if _backend == "tensorflow":
                return lambda x: _tf.convert_to_tensor(x) if isinstance(x, (list, tuple)) else _tf.expand_dims(x, axis = 0)
            
            else:
                return _np.atleast_1d
        
        # (X): Return the computed attribute according to the backend setting:
        return getattr(mod, name)
    
    def safe_cast(self, x, promote_to_complex_if_needed = False):
        """
        ## Description:
        For a description of what's happening, please see the function `safe_cast` above.
        """
        return safe_cast(x, promote_to_complex_if_needed)
    
    def promote_scalar_to_dtype(self, scalar, reference):
        """
        ## Description:
        Promote a Python scalar to the same dtype as the reference tensor/array.
        Works under both NumPy and TensorFlow backends.

        ## Arguments:
            scalar: A Python scalar (float or int)

            reference: A tensor or array whose dtype will be matched

        ## Returns:
            A scalar of the same backend and dtype as `reference`
        """

        if _backend == "tensorflow":
            if hasattr(reference, "dtype"):
                ref_dtype = reference.dtype
            else:
                ref_dtype = _tf.complex64 if isinstance(reference, complex) else _tf.float32

            if ref_dtype.is_complex:
                return _tf.complex(
                    _tf.constant(scalar, dtype = _tf.float32),
                    _tf.constant(0.0, dtype = _tf.float32)
                )
            else:
                return _tf.constant(scalar, dtype=ref_dtype)

        elif _backend == "numpy":
            reference_datatype = reference.dtype if hasattr(reference, "dtype") else _np.float32

            return _np.array(scalar, dtype = reference_datatype)

        else:
            raise ValueError(f"Unsupported backend: {_backend}")
        
# (X): Export the wrapper that handles the library-specific attribute business:
math = MathWrapper()



"""
A testing suite for proving that the functions computing the coefficients for the unpolarized
target are returning real, finite, and accurate values.
"""

# (X): Native Library | unittest:
import unittest

# (X): External Library | NumPy:
import numpy as np

# (X): Self-Import | BKM10Inputs:
from bkm10_lib.inputs import BKM10Inputs

# (X): Self-Import | CFFInputs:
from bkm10_lib.cff_inputs import CFFInputs

# (X): Self-Import | DifferentialCrossSection
from bkm10_lib.core import DifferentialCrossSection

# (X): Self-Import | BKMFormalism
from bkm10_lib.formalism import BKMFormalism


# (X): Define a class that inherits unittest's TestCase:
class TestUnpolarizedCoefficients(unittest.TestCase):

    TEST_LAB_K = 5.75
    TEST_Q_SQUARED = 1.82
    TEST_X_BJORKEN = 0.343
    TEST_T = -0.172

    @classmethod
    def setUpClass(cls):

        # (X): Provide the BKM10 inputs to the dataclass:
        cls.test_kinematics = BKM10Inputs(
            lab_kinematics_k = cls.TEST_LAB_K,
            squared_Q_momentum_transfer = cls.TEST_Q_SQUARED,
            x_Bjorken = cls.TEST_X_BJORKEN,
            squared_hadronic_momentum_transfer_t = cls.TEST_T)

        # (X): Provide the CFF inputs to the dataclass:
        cls.test_cff_inputs = CFFInputs(
            compton_form_factor_h = complex(-0.897, 2.421),
            compton_form_factor_h_tilde = complex(2.444, 1.131),
            compton_form_factor_e = complex(-0.541, 0.903),
            compton_form_factor_e_tilde = complex(2.207, 5.383))
        
        # (X): Specify the target polarization *as a float*:
        cls.target_polarization = 0.

        # (X): Specify the beam polarization *as a float*:
        cls.lepton_polarization = 0.0

        # (X): We are using the WW relations in this computation:
        cls.ww_setting = True

        # (X): Using the setting we wrote earlier, we now need to construct
        cls.configuration = {
            "kinematics": cls.test_kinematics,
            "cff_inputs": cls.test_cff_inputs,
            "target_polarization": cls.target_polarization,
            "lepton_beam_polarization": cls.lepton_polarization,
            "using_ww": cls.ww_setting
        }
        
        cls.cross_section = DifferentialCrossSection(configuration = cls.configuration)

        cls.phi_values = np.linspace(0, 2 * np.pi, 10)
    
    def assert_is_finite(self, value):
        """
        ## Description:
        A general test in the suite that verifies that all the
        numbers in an array are *finite* (as opposed to Inf.-type or NaN)

        ## Notes:
        "NaN" means "not a number." Having NaN values in an array causes problems in
        functions that are designed to perform arithmetic.
        """
        self.assertTrue(
            expr = np.isfinite(value).all(),
            msg = "Value contains NaNs or infinities/Inf.")

    def assert_no_nans(self, value):
        """
        ## Description:
        A general test in the suite that determines if an array has NaNs.
        
        ## Notes:
        "NaN" means "not a number." Having NaN values in an array causes problems in
        functions that are designed to perform arithmetic.
        """
        self.assertFalse(
            expr = np.isnan(value).any(),
            msg = "> [ERROR]: Value contains NaNs")

    def assert_no_negatives(self, value):
        """
        ## Description:
        A general test in the suite that determines if an array has negative values
        in it.

        ## Notes:
        There *are* important negative quantities, and several coefficients are indeed
        negative. But cross-sections, for example, should be positive.
        """
        self.assertTrue(
            expr = (value >= 0).all(),
            msg = "> [ERROR]: Value contains negative values")

    def assert_is_real(self, value):
        """
        ## Description:
        A general test in the suite that determines that an array has
        all real values.
        """
        self.assertTrue(
            expr = np.isreal(value).all(),
            msg = "> [ERROR]: Value contains complex components")

    def assert_approximately_equal(self, value, expected, tolerance = 1e-8):
        """
        ## Description:
        A general test in the suite that determines if a *number* (`value`) is approximately
        equal to what is expected (`expected`). "Approximately equal" is quantified with the 
        parameter `tolerance`.
        """
        self.assertTrue(
            np.allclose(value, expected, rtol = tolerance, atol = tolerance),
            f"> [ERROR]: Expected {expected}, got {value}")
        
    def test_c0_coefficient(self):
        """
        ## Description:
        Test one of the highest-level coefficients in the BKM10 formalism:
        c_{0}^{I} in the mode expansion.
        """

        # (X): Calculate c_{0}^{I}:
        c0 = self.cross_section.compute_c0_coefficient(self.phi_values)[0]

        # (X): Verify that c_{0}^{I} is a *finite* number:
        self.assert_is_finite(c0)
        
        # (X); Verify that c_{0}^{I} is not a NaN:
        self.assert_no_nans(c0)

        # (X): Verify that c_{0}^{I} is real:
        self.assert_is_real(c0)

        # (X): IMPORTANT ONE: Verify that c_{0}^{I} is what we expect:
        _MATHEMATICA_RESULT = 4.196441097163937 + 29.512298473681934 - 0.4548568231402324
        self.assert_approximately_equal(c0, expected = _MATHEMATICA_RESULT)

    def test_c1_coefficient(self):
        """
        ## Description:
        Test one of the highest-level coefficients in the BKM10 formalism:
        c_{1}^{I} in the mode expansion.
        """

        # (X): Calculate c_{1}^{I}:
        c1 = self.cross_section.compute_c1_coefficient(self.phi_values)[0]

        # (X): Verify that c_{1}^{I} is a *finite* number:
        self.assert_is_finite(c1)
        
        # (X); Verify that c_{1}^{I} is not a NaN:
        self.assert_no_nans(c1)

        # (X): Verify that c_{1}^{I} is real:
        self.assert_is_real(c1)

        # (X): IMPORTANT ONE: Verify that c_{1}^{I} is what we expect:
        _MATHEMATICA_RESULT = -0.3460689391000681
        self.assert_approximately_equal(c1, expected = _MATHEMATICA_RESULT)

    def test_c2_coefficient(self):
        """
        ## Description:
        Test one of the highest-level coefficients in the BKM10 formalism:
        c_{2}^{I} in the mode expansion.
        """

        # (X): Calculate c_{2}^{I}:
        c2 = self.cross_section.compute_c2_coefficient(self.phi_values)[0]

        # (X): Verify that c_{2}^{I} is a *finite* number:
        self.assert_is_finite(c2)
        
        # (X); Verify that c_{2}^{I} is not a NaN:
        self.assert_no_nans(c2)

        # (X): Verify that c_{2}^{I} is real:
        self.assert_is_real(c2)

        # (X): IMPORTANT ONE: Verify that c_{2}^{I} is what we expect:
        _MATHEMATICA_RESULT = -0.03259012849881058
        self.assert_approximately_equal(c2, expected = _MATHEMATICA_RESULT)

    def test_c3_coefficient(self):
        """
        ## Description:
        Test one of the highest-level coefficients in the BKM10 formalism:
        c_{3}^{I} in the mode expansion.
        """

        # (X): Calculate c_{3}^{I}:
        c3 = self.cross_section.compute_c3_coefficient(self.phi_values)[0]

        # (X): Verify that c_{2}^{I} is a *finite* number:
        self.assert_is_finite(c3)
        
        # (X); Verify that c_{2}^{I} is not a NaN:
        self.assert_no_nans(c3)

        # (X): Verify that c_{2}^{I} is real:
        self.assert_is_real(c3)

        # (X): IMPORTANT ONE: Verify that c_{2}^{I} is what we expect:
        _MATHEMATICA_RESULT = 0.0003562823963322977
        self.assert_approximately_equal(c3, expected = _MATHEMATICA_RESULT)


if __name__ == "__main__":
    unittest.main()


"""
The entry point for the dataclass `CFFInputs`
"""

# (1): Import the specialized `dataclass` library:
from dataclasses import dataclass

# (2): 
from bkm10_lib import backend

# (3): Immediately define the CFFInputs dataclass:
@dataclass
class CFFInputs:
    """
    Welcome to the `CFFInputs` dataclass!

    ## Description:
    The CFFs (Compton Form Factors) live inside the BKM10 cross section.
    They are unknown functions that also parametrize the DVCS process. Please
    read the "notes" section below for motivation behind why we use it. What is
    relevant for the programmer here is that these CFFs *must* be supplied in 
    order to return a cross section. You need to make at least a *guess* for
    what their values are.

    ## Notes:
    The CFFs enter the cross section through loop integration of a perturbativley hard 
    interaction and a more general structure function called a GPD (generalized parton
    distribution) over the parton momentum fraction (which we call x). A lot of this 
    business of measuring DVCS cross sections is really about figuring out the form of
    the CFFs and the GPDs. However, in order to make predictions about the cross-section,
    these functions require input themselves.
    """
    
    # (1): The CFF H --- Requires Re[H] and Im[H], so it's of `complex` type:
    compton_form_factor_h: complex

    # (2): The CFF H_tilde --- Requires Re[Ht] and Im[Ht]:
    compton_form_factor_h_tilde: complex

    # (3): The CFF E --- Requires Re[E] and Im[E]:
    compton_form_factor_e: complex

    # (4): The CFF E_tilde --- Requires Re[Et] and Im[Et]:
    compton_form_factor_e_tilde: complex

    def conjugate(self):
        """
        ## Description:
        Computes the complex conjugate each of the CFFs in this dataclasss.
        """

        conjugate_function = backend.math.conj

        return CFFInputs(
            compton_form_factor_h = conjugate_function(self.compton_form_factor_h),
            compton_form_factor_h_tilde = conjugate_function(self.compton_form_factor_h_tilde),
            compton_form_factor_e = conjugate_function(self.compton_form_factor_e),
            compton_form_factor_e_tilde = conjugate_function(self.compton_form_factor_e_tilde),
        )



"""
Entry point for installing the package.
"""

# Native Library | setuptools
from setuptools import setup, find_packages

# (1): Use setup() to... set up the package:
setup(
    name = "bkm10_lib",
    version = "1.3.2",
    description = "A Python library to help nuclear physicists use the BKM formalism in predicting cross-section, asymmetries, and comparing GPD models.",
    author = "Woofmagic",
    author_email = "none@none.none",
    url = "https://github.com/Woofmagic/bkm10_lib",
    project_urls = {
        "Sources": "https://github.com/Woofmagic/bkm10_lib",
        "Bug Tracker": "https://github.com/Woofmagic/bkm10_lib/issues",
        'Changelog': 'https://github.com/Woofmagic/bkm10_lib/CHANGELOG.rst',
    },
    packages = find_packages(),
    classifiers = [
        # (X): For the complete classifier list, visit: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 2 - Pre-Alpha', # There are 7 of these --- make sure you know which one is correct!
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Physics',
        'Natural Language :: English'
    ],
    keywords = [
        'Physics', 
        'Nuclear Physics', 
        'Particle Physics',
        'Hadronic Physics',
        'Form Factors',
        'Parton Distributions',
        'Parton Distribution Functions',
        'Generalized Parton Distributions',
        'NumPy',
        'Pandas',
        'TensorFlow',
        'Mathematics',
    ],
    python_requires = '>=3.7',
)
